#!/usr/bin/env python3

from pathlib import Path

def main():
    def input_files():
        for path in Path('.').glob('*.*'):
            if path.suffix in ('.html', '.txt'):
                yield path
        
    for path in input_files():
        text = path.read_text(encoding='utf8')

        dstpath = path.parent / (path.name + '.utf16')
        print("Writing %s" % dstpath)
        dstpath.write_text(text, encoding='utf16')


if __name__ == '__main__':
    main()



import json, sys

with open(sys.argv[1] + "/results.json", 'rt') as f:
    data = json.load(f)

tests = data['tests']
solutions = data['solutions']

print('<tr>')
print('  <th></th>')
for sol in solutions:
    print('  <th>%s</th>' % sol)
print('</tr>')
for test in tests:
    print('<tr>')
    print('  <th>%s</th>' % test)
    for sol in solutions:
        values = data['xdata'][test][sol]
        for i in range(2):
            if values[i] is None:
                values[i] = '???'
            else:
                values[i] = '%.2f' % values[i]
        print('  <td>%s / %s</td>' % (values[0], values[1]))
    print('</tr>')



## prefix_to_length_table
## We do
## const uint8_t prefix = (input[position] >> 5);
## so
## 00000000 becomes  000
## 10000000 becomes  100
## 11000000 becomes  110
## 11100000 becomes  111
prefix_to_length_table = []
for i in range(0b111 + 1):
    if(i < 0b100):
        # ascii
        prefix_to_length_table.append(1)
    elif(i<0b110):
        # continuation
        prefix_to_length_table.append(0)
    elif(i<0b111):
        prefix_to_length_table.append(2)
    else:
        prefix_to_length_table.append(3)
print("prefix_to_length_table")
print(prefix_to_length_table)


def decode(i):
    answer = []
    for j in range(8):
        rem = (i % 3) + 1
        answer.append(rem)
        i = i // 3
    answer.reverse()
    return answer

table_pattern1=[]
table_pattern2=[]

for i in range(3**8):
    x = decode(i)
    # pattern1 captures the second and third position
    # pattern2 captures the first position
    pattern1 = []
    pattern2 = []
    pos = 0
    for i in x:
        pattern2.append(pos+i-1)
        pattern2.append(0xFF)
        if(i == 1):
            pattern1.append(0xFF)
            pattern1.append(0xFF)
        elif(i == 2):
            pattern1.append(pos)
            pattern1.append(0xFF)
        elif(i==3):
            pattern1.append(pos +1)
            pattern1.append(pos)
        else:
            print("BUG")
        pos += i
    table_pattern1.append(pattern1)
    table_pattern2.append(pattern2)
    assert(len(pattern1) == 16)
    assert(len(pattern2) == 16)

print("const static uint8_t pattern1["+str(len(table_pattern1))+"][16]={")
for x in table_pattern1:
    assert(len(x) == 16)
    print('{%s},' % (', '.join(map(str, x))))
print("};")
print("const static uint8_t pattern2["+str(len(table_pattern1))+"][16]={")
for x in table_pattern2:
    assert(len(x) == 16)
    print('{%s},' % (', '.join(map(str, x))))
print("};")



# 
#  make_simd_operation_test.py
#
#  Copyright (C) 2007 Dan Lin, Robert D. Cameron
#  Licensed to International Characters Inc. and Simon Fraser University
#	              under the Academic Free License version 3.0
#  Licensed to the public under the Open Software License version 3.0.

# make_simd_operation_test.py generates test file simd_operation_test.c
# which compares the result computed by simd operations and 
# result simulated by this test file

from random import randint

ops = ["rotl"]
mods = ["x","h","l"]
fws = [32]
merge_fws = fws[:-1]
bitBlock_size = 64
r1=0
r2=0

def modified_operand(N, n, operand, modifier):
    if modifier == "x": return operand
    elif modifier == "h":
        for i in range(0,N/n):
            operand[i] = operand[i] >> ((n+1)/2)
        return operand
    else:
        for i in range(0,N/n):
            operand[i] = operand[i] % (2** ((n+1)/2))
        return operand

def split_up(r,N,n):
    remainder = r
    result = []
    for i in range(0,N/n):
        quotient = remainder / (2**(N-n*(i+1)))
        remainder = remainder % (2**(N-n*(i+1)))
        result.append(quotient)
    return result

def join_up(r,N,n):
    result = 0
    for i in range(0,N/n):
        result = result + (r[i] * (2**(N-n*(i+1))))
    return result

def gen_const(c,N,n):
    return c*(2**N-1)/(2**n-1)
 
def get_mask(n):
    count = 1
    temp = n
    while (n/2 != 1):
        n = n/2
        count = count + 1
    return count

def simulate_add (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        r3_field.append((r1_field[i] + r2_field[i])%(2**n))
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_sub (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        r3_field.append((r1_field[i] - r2_field[i] + 2**n)%(2**n))
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_srl (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        r3_field.append(r1_field[i] >> (r2_field[i] % 2**get_mask(n)))
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_sll (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        r3_field.append((r1_field[i] << (r2_field[i] % 2**get_mask(n)))% (2**n))
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_sra (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        supplement = 0
        shift = r2_field[i] % (2**get_mask(n))
        for j in range(1, shift+1):
            supplement = supplement + (r1_field[i]/(2**(n-1)))*2**(n-j)
        r3_field.append((r1_field[i] >> shift) + supplement)
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_rotl (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        shift = r2_field[i] % 2**get_mask(n)
        r3_field.append((r1_field[i] << shift) % 2**n + ((r1_field[i] >> (n - shift))% (2**n)))
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_mergeh (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n/2):
        r3_field.append(r1_field[i])
        r3_field.append(r2_field[i])
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_mergel (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(N/n/2, N/n):
        r3_field.append(r1_field[i])
        r3_field.append(r2_field[i])
    r3 = join_up(r3_field, N, n)
    return r3

def simulate_pack (N, n, m1, m2, r1, r2):
    r1_field = modified_operand(N, n, split_up(r1,N,n), m1)
    r2_field = modified_operand(N, n, split_up(r2,N,n), m2)
    r3_field = []
    for i in range(0, N/n):
        r3_field.append(r1_field[i] % (2** (n/2)))
    for j in range(0, N/n):
        r3_field.append(r2_field[j] % (2** (n/2)))
    r3 = join_up(r3_field, N, n/2)
    return r3
 
def make_modified_operand(fw, operand, modifier):
    if modifier == "x": return operand
    elif modifier == "h": return "simd_srli_"+str(fw)+"(%s, %i)" % (operand, fw/2)
    else: return "simd_andc(%s, simd_himask_%i)" % (operand, fw)

def split_64to16(r):
    result = []
    result_hi =  r/2**32
    result_lo =  r%2**32
    result.append(result_hi/2**16)
    result.append(result_hi%2**16)
    result.append(result_lo/2**16)
    result.append(result_lo%2**16)
    return result

def reverse(n):
    mask = 255
    result = 0
    for i in range(0,8):
	    temp = (n & mask)>>(i*8)
	    result = (result | temp)<< 8
	    mask = mask << 8
    return result>>8
    
def make_test(op, fw, m1, m2, r1, r2):
    test_operation = "simd_%s_%i_%s%s (a, b)"
    template1 = "\ttest_rslt = %s;\n" % test_operation
    template1 += "\tif (!(simd_all_eq_8(test_rslt, simd_if(simd_himask_64,simd_if(simd_himask_32,simd_const_16(%s),simd_const_16(%s)),simd_if(simd_himask_32,simd_const_16(%s),simd_const_16(%s)))))) {\n"
    err_stmts = "\t\t\tprintf(\"error in %s \\n\");\n" % test_operation
    err_stmts += "\t\t\tprint_bit_block(\"Computed:\", test_rslt);\n"
    err_stmts += "\t\t\tprintf(\"\\tExpected: %s\\n\");\n"
    err_stmts += "\t\t}\n"
    template = template1 + err_stmts
    if op == "add":
	r =  simulate_add(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "sub":
	r =  simulate_sub(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "pack":
	r =  simulate_pack(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "mergel":
	r =  simulate_mergel(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "mergeh":
	r =  simulate_mergeh(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "srl":
	r =  simulate_srl(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "sll":
	r =  simulate_sll(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "sra":
	r =  simulate_sra(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    elif op == "rotl":
	r =  simulate_rotl(bitBlock_size,fw,m1,m2,r1,r2)
	result = split_64to16(r)
        return template % (op, fw, m1, m2,result[0],result[1],result[2],result[3],op,fw,m1,m2,hex(reverse(r)))
    else:
        return template % (op, fw, m1, m2, "0x0","0x0","0x0","0x0",op,fw,m1,m2,"0x0")


def make_all_for_fw(fw,r1,r2):
    test_list = ''
    for m1 in mods:
        for m2 in mods:
            for op in ops:
                test_list += make_test(op, fw, m1, m2, r1, r2)
    return test_list

def gen_operand_sequence(fw):
    if fw == 2:
	test_operand = [0x0, 0x1, 0x2, 0x3]
    elif fw == 4:
	test_operand = [0x0, 0x7, 0x8, 0xf]
	test_operand.append(randint(0x1, 0xe))
    elif fw == 8:
	test_operand = [0x0, 0x7f, 0x80, 0xff]
	test_operand.append(randint(0x1, 0xfe))
    elif fw == 16:
	test_operand = [0x0, 0x7fff, 0x8000, 0xffff]
	test_operand.append(randint(0x1, 0xfffe))
    elif fw == 32:
	test_operand = [0x0, 0x7fffffff, 0x80000000, 0xffffffff]
	test_operand.append(randint(0x1, 0xfffffffe))
    return test_operand


def generate_and_write_versions(filename, ops,fws):
    test_list = r"""#include <stdio.h>
#include "%s.h"
int main() {
        SIMD_type a, b,test_rslt;
"""
    test_list = test_list % filename
    for fw in fws:      
	for r1 in gen_operand_sequence(fw):
	      for r2 in  gen_operand_sequence(fw):
                  test_list +='\ta=simd_const_'+str(fw)+'('+hex(r1)+');\n\tb=simd_const_'+str(fw)+'('+hex(r2)+');\n'
                  test_list += make_all_for_fw (fw,gen_const(r1,64,fw),gen_const(r2,64,fw))
    test_list += '\treturn 0;\n}\n'
    file = open(filename+"_test.c", 'w')
    file.write(test_list)
    file.close()
    
import sys
if __name__ == "__main__":
    if len(sys.argv) < 2: library_under_test = "mmx_simd"
    else: library_under_test = sys.argv[1]
    generate_and_write_versions(library_under_test, ops, fws)



def format(number):
    upper = number // (1 << 64)
    lower = number % (1 << 64)
    print("" + hex(upper) + "," + hex(lower) + ",")


for q in range(-342, 0):
    power5 = 5 ** -q
    z = 0
    while (1 << z) < power5:
        z += 1
    if q >= -27:
        b = z + 127
        c = 2 ** b // power5 + 1
        format(c)
    else:
        b = 2 * z + 2 * 64
        c = 2 ** b // power5 + 1
        # truncate
        while c >= (1 << 128):
            c //= 2
        format(c)

for q in range(0, 308 + 1):
    power5 = 5 ** q
    # move the most significant bit in position
    while power5 < (1 << 127):
        power5 *= 2
    # *truncate*
    while power5 >= (1 << 128):
        power5 //= 2
    format(power5)



# 
#  make_half_operand_versions.py
#
#  Copyright (C) 2007 Robert D. Cameron, Dan Lin
#  Licensed to International Characters Inc. and Simon Fraser University
#	              under the Academic Free License version 3.0
#  Licensed to the public under the Open Software License version 3.0.
#
# make_halfoperand_versions.py generates macro definitions for
# the half-operand versions of idealized SIMD operations.
#
# The half-operand versions of a (binary) operation on n-bit fields
# are named by # extending the operation name with an underscore and 
# two # operand modification letters for the two operands:
#   operand modifier "h" indicates that the high n/2 bits of fields
#      are taken as the operand values (shifted right by n/2).
#   operand modifier "l" indicates that the low n/2 bits of fields
#      masking off the high n/2 bits.
#   operand modifier "x" indicates that all n bits of the field are
#      used as the operand value without modification.
#
ops = ["sub", "add", "pack", "mergeh", "mergel", "sll", "srl", "sra", "rotl"]
mods = ["x", "l", "h"]
fws = [2, 4, 8, 16, 32]             

# The generic approach to defining these operations is to simply
# include code to modify each of the operands and then perform
# the desired operation.  However, this program generates more
# efficient code when certain optimizations may apply.
#
# 1.  Lo-Mask Optimization.
#     The masking for an "l" operand can be eliminated if
#     that operation ignores the high n/2 bits.  The
#     ignores_operand_hi_half property declares this.
# 2.  Hi-Mask Optimization
#     The shifting of an "h" operand can be optimized if the
#     operation ignores the high n/2 bits of the operand.  In this 
#     case, it is possible to replace a simulated shift for n bit
#     fields with any built-in shift for fields of a multiple of n.
# 3.  Double-High Shift Optimization
#     In some cases, the two shift operations for a pair of "h" operands 
#     may be eliminated in favor of a single shift operation after
#     the basic operation is perfomed (support_hh_postshift property).
# 4.  LH Add Optimization
#     If both operands are modified (either "h" or "l" modifiers), 
#     a simulated add on small field widths can be replaced by a
#     built-in add operation for fields of a multiple of n.
#
##########################################################################
#
# Operation Properties to Support Optimization
#
# 1.  Lo-Mask Optimization.
#
# An operation satisfies the ignores_operand_hi_half property
# for a particular operand if the high n/2 bits of that
# operand play no role in the operation applied to n bit
# fields.   This property applies for the second operand
# of all shifts and rotates, as well as for both operands
# of packs.
#
ignores_operand_hi_half = { 1: ["pack"], 2: ["sll", "srl", "sra", "rotl", "pack"]}
#
# 2.  Hi-Mask Optimization
#
# Shifts of 2, 4 or 8 bits are simulated on common architectures.  Replace
# if possible with  16-bit shifts, which are built-in on common architectures.
#  
simulated_shift_replacement = {2: 16, 4: 16, 8: 16}
#
# 3.  Double-High Shift Optimization
#
support_hh_postshift = ["mergeh", "mergel"]
#
# 4.  LH Add Optimization
#
# Adds of 2-bit or 4-bit fields are simulated on common architectures.  Replace
# if possible with 8-bit adds, which are built-in on common architectrues.
#
simulated_add_replacement = {2: 8, 4: 8}

#
# Generate a definition that won't override any hand-coded version.
#
defn_mode = "INLINE"

def define_macro_if_undefined(op_name, body):
    return "#ifndef %s\n#define %s(v1, v2) %s\n#endif\n\n" % (op_name, op_name, body)

def define_inline_function_if_undefined(op_name, body):
    prototype = "inline SIMD_type %s(SIMD_type v1, SIMD_type v2)" % (op_name)
    return "#ifndef %s\n%s {\n  return %s;\n}\n#endif\n\n" % (op_name, prototype, body)

def define_if_undefined(op_name, body):
    global defn_mode
    if defn_mode == "MACRO": return define_macro_if_undefined(op_name, body)
    else: return define_inline_function_if_undefined(op_name, body)

def operand_name(operand_no):
    return "v%i" % operand_no

def make_modified_operand(operation, fw, operand_no, modifier):
    operand = operand_name(operand_no)
    if modifier == "x": return operand
    elif operation in ignores_operand_hi_half[operand_no]:
	if modifier == "l": return operand
	elif modifier == "h":
	    if fw in simulated_shift_replacement.keys():
	      shft_op = "simd_srli_%i" % (simulated_shift_replacement[fw])
	    else: shft_op = "simd_srli_%i" % (fw)
	    return "%s(%s, %i)" % (shft_op, operand, fw/2)
	else: raise Exception("Bad modifier %s" % modifier)
    elif modifier == "h": return "simd_srli_%i(%s, %i)" % (fw, operand, fw/2)
    elif modifier == "l": return "simd_andc(%s, simd_himask_%i)" % (operand, fw)
    else: raise Exception("Bad modifier %s" % modifier)


def make_optimized_defn(op, fw, m1, m2):
    base_operation = "simd_%s_%i" % (op, fw)
    op_name = base_operation + "_" + m1 + m2
    operand1 = make_modified_operand(op, fw, 1, m1)
    operand2 = make_modified_operand(op, fw, 2, m2)
    if (m1 == "h") and (m2 == "h") and (op in support_hh_postshift):
        code = "%s(%s, %s)" % (base_operation, operand1, operand2)
	return define_if_undefined(op_name, code)
    if (op == "add") and (m1 != "x") and (m2 != "x") and (fw in simulated_add_replacement.keys()):
	base_operation = "simd_%s_%i" % (op, simulated_add_replacement[fw])
    return define_if_undefined(op_name, "%s(%s, %s)" % (base_operation, operand1, operand2))

def make_all_for_op_fw(op, fw):
    defn_list = ''
    for m1 in mods:
        for m2 in mods:
            defn_list += make_optimized_defn(op, fw, m1, m2)
    return defn_list

#
# Usage: make_all_half_operand_versions(ops, fws) to generate
# a complete file of all the "half-operand" modified versions
# of a set of simd operations for each operation op within
# the list ops and each field width fw in the list of field
# widths.
#
def make_all_half_operand_versions (ops, fws):
    defn_list = ''
    for op in ops:
        for fw in fws:
            defn_list += make_all_for_op_fw(op, fw)
    return defn_list

def generate_and_write_versions(filename, ops, fws):
    defn_list = make_all_half_operand_versions (ops, fws)
    file = open(filename, 'w')
    file.write(defn_list)
    file.close()
    
if __name__ == "__main__":
    generate_and_write_versions("mmx_simd_modified.h", ops, fws)   



import subprocess, os, sys, re, shutil, json

RunsCount = 10000

ConvertToUtf16 = "FileConverter_msvc %s %s"

Solutions_Decode = {
    "trivial":    "FileConverter_msvc %s temp.out -b=0 -k={0} --small",
    "utf8lut:1S": "FileConverter_msvc %s temp.out -b=3 -k={0} --small",
    "utf8lut:4S": "FileConverter_msvc %s temp.out -b=3 -k={0}",
    "u8u16": "u8u16_ssse3 %s temp.out {0}",
    "utf8sse4": "utf8sse4 %s temp.out {0}",
}
Solutions_Encode = {
    "trivial":    "FileConverter_msvc -s=utf16 -d=utf8 %s temp.out -b=0 -k={0} --small",
    "utf8lut:1S": "FileConverter_msvc -s=utf16 -d=utf8 %s temp.out -b=3 -k={0} --small",
    "utf8lut:4S": "FileConverter_msvc -s=utf16 -d=utf8 %s temp.out -b=3 -k={0}",
}

encode = (sys.argv[1] == 'encode')
Solutions = (Solutions_Encode if encode else Solutions_Decode)
for k,v in Solutions.items():
    Solutions[k] = v.format(RunsCount)

Tests = {
    "[rnd1111:400000]": "rnd1111_utf8.txt",
    "[rnd1110:500000]": "rnd1110_utf8.txt",
    "chinese": "chinese_book.txt",
    "russian": "war_and_piece.fb2",
    "english": "english_book.txt",
    "unicode": "unicode_table.html",
}

def log_name_of(sol_name, test_name):
    log_name = sol_name + "__" + test_name
    log_name = "logs/" + re.sub(r'[\W]', '_', log_name) + ".log"
    return log_name

def run_sol_test(sol_name, test_name, encode=False):
    if encode:
        subprocess.run((ConvertToUtf16 % (Tests[test_name], 'temp.in')).split(), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        cmd = Solutions[sol_name] % 'temp.in'
    else:
        cmd = Solutions[sol_name] % Tests[test_name]
    log_name = log_name_of(sol_name, test_name)
    with open(log_name, 'wt') as f:
        subprocess.run(cmd.split(), stdout=f, stderr=f)
    return log_name

def parse_log(log_name):
    with open(log_name, 'rt') as f:
        text = f.read()
    external_result = None
    internal_result = None
    # typical overall measurement, added to every solution by myself
    m = re.search(r'From total time\s+:\s+([\d.]+)\s+cyc/el\n', text)
    if m:
        external_result = float(m.group(1))
    # internal timings in utf8lut
    for m in re.finditer(r'slot \d+\s+(DECODE|ENCODE)\s+:\s+([\d.]+) cyc/el\s+(\d+) elems\n', text):
        if int(m.group(3)) > 0:
            internal_result = float(m.group(2))
    # internal timings in u8u16
    for m in re.finditer(r'BOM \d+: \d+ \(avg time: \d+ cyc/kElem\) Cumulative: \d+ \(avg: (\d+) cyc/kElem\)', text):
        internal_result = float(m.group(1)) / 1000.0
    return internal_result, external_result

def main():
    shutil.rmtree('logs', ignore_errors=True)
    os.mkdir('logs')

    #print(parse_log(run_sol_test("utf8lut:1S", "chinese")))
    #print(parse_log(run_sol_test("u8u16", "chinese")))
    #print(parse_log(run_sol_test("utf8sse4", "chinese")))

    jsonall = {}
    jsonall['solutions'] = list(Solutions.keys())
    jsonall['tests'] = list(Tests.keys())
    data = jsonall['xdata'] = {}
    for test in Tests:
        data[test] = {}
        for sol in Solutions:
            print("Test %s, Sol %s:" % (test, sol), end="", flush=True)
            log_name = run_sol_test(sol, test, encode)
            print(" finished: ", end="", flush=True)
            result = parse_log(log_name)
            print(result[0], '/', result[1], flush=True)
            data[test][sol] = (result[0], result[1], log_name)

    with open("logs/results.json", "wt") as f:
        json.dump(jsonall, f, indent=2, sort_keys=True)

if __name__ == "__main__":
    main()



# 
#  make_basic_ops.py
#
#  Copyright (C) 2007 Dan Lin, Robert D. Cameron
#  Licensed to International Characters Inc. and Simon Fraser University
#	              under the Academic Free License version 3.0
#  Licensed to the public under the Open Software License version 3.0.

# make_basic_ops.py generates inline definitions for
# idealized basic SIMD operations that are not included
# in the built-in functions.
ops = ["sub", "add", "sll", "srl", "sra", "pack", "mergeh", "mergel", "rotl"]

# 1 for "add" and "sub"
# 2 for "sll", "srl" and "sra"
# 3 for "pack"
# 4 for "mergel" and "mergeh"
# Note: For new operations added into this file,
#       if different fieldwidths are needed, add more to the following list
#       otherwise share with the operations above
fws = { 1: [4], 2: [32,16,8,4], 3: [8,4,2], 4: [4,2,1]}

ops_immediateshift= ["sll","srl"]
fws_immediateshift = [2,4,8]


# make the structure of a inline function

def make_inline(op,fw,body):
	operation = "simd_%s_%i" % (op,fw)
	return "#ifndef %s\ninline SIMD_type %s(SIMD_type a,SIMD_type b){\n\treturn %s;}\n#endif\n\n" % (operation,operation,body)

# this is the main function that generate simd operations in n bit field
# by using simd operations in 2n bit field
#
# Considering that simd_const_64() is not included in simd_built_in operations
# we use _mm_cvtsi32_si64() as the mask for shift operations in 32 bit field
# instead of masking the operand twice with simd_himask_32 and smid_const_32()
#
# Note: "rotl" has not been tested
#       "srl", "sll" and "sra" in 4 bit field might not be efficient
#       by using 8 bit field operations
def make_halfsize_defn(op,fw):
    template = "simd_if(simd_himask_"+str(fw*2)+", simd_"+op+"_%i(%s,%s)\n\t,"+"simd_"+op+"_%i(%s,%s))"
    if (op == "add") or (op == "sub"):
        return template % (fw*2, "simd_and(a,simd_himask_"+str(fw*2)+")",
    			  "simd_and(b,simd_himask_"+str(fw*2)+")",
			  fw*2,"simd_andc(a,simd_himask_"+str(fw*2)+")",
			  "simd_andc(b,simd_himask_"+str(fw*2)+")")
    elif op == "srl":
	common_part = template % (fw*2,"a","simd_and(simd_const_"+str(fw)+"("+str(fw-1)+"),simd_srli_"+str(fw*2)+"(b,"+str(fw)+"))",
				  fw*2,"simd_andc(a,simd_himask_"+str(fw*2)+")","simd_and(%s)")
	if fw == 32:
		return common_part % "_mm_cvtsi32_si64(31),b"
	else:
		return common_part % ("b,simd_const_"+str(fw*2)+"("+str(fw-1)+")")
    elif op == "sll":
	common_part = template % (fw*2, "simd_and(a,simd_himask_"+str(fw*2)+")",
				  "simd_and(simd_const_"+str(fw)+"("+str(fw-1)+"),simd_srli_"+str(fw*2)+"(b,"+str(fw)+"))",
				  fw*2,"a", "simd_and(%s)")
	if fw == 32:
		return common_part % "_mm_cvtsi32_si64(31),b"
	else:
		return common_part % ("b,simd_const_"+str(fw*2)+"("+str(fw-1)+")")
    elif op == "sra":
	if fw == 32:
		return """simd_if(simd_himask_64(),
			_mm_sra_pi32(a, simd_and(simd_const_32(31),simd_srli_64(b,32))),
			_mm_sra_pi32(a, simd_and(_mm_cvtsi32_si64(31), b)))"""
        else:		
		return template % (fw*2,"simd_and(a,simd_himask_"+str(fw*2)+")",
				  "simd_and(sisd_srli(b,"+str(fw)+"),simd_const_"+str(fw*2)+"("+str(fw-1)+"))",
				  fw*2,"simd_srai_"+str(fw*2)+"(sisd_slli(a,"+str(fw)+"),"+str(fw)+")",
				  "simd_and(b,simd_const_"+str(fw*2)+"("+str(fw-1)+"))")
    elif op == "rotl":
        return "simd_or(simd_sll_"+str(fw)+"(a,b),simd_srl_"+str(fw)+"(a,simd_sub_"+str(fw)+"(simd_const_"+str(fw)+"("+str(fw)+"),b)))"
    elif op == "pack":
	return "simd_pack_"+str(fw*2)+"(%s,\n\t%s)" % ("simd_if(simd_himask_"+str(fw)+",sisd_srli(a,"+str(fw/2)+"),a)",
    							"simd_if(simd_himask_"+str(fw)+",sisd_srli(b,"+str(fw/2)+"),b)")
    elif op == "mergeh" or op == "mergel":
	return "simd_"+op+"_"+str(fw*2)+"(%s,\n\t%s)" % ("simd_if(simd_himask_"+str(fw*2)+",a,sisd_srli(b,"+str(fw)+"))",
    							"simd_if(simd_himask_"+str(fw*2)+",sisd_slli(a,"+str(fw)+"),b)")
    else:
	raise Exception("Bad operator %s" % op)

def make_immediateshift_defn(op,fw):
    template = "inline SIMD_type simd_"+op+"i_"+str(fw)+"(SIMD_type r, int sh){\n\t return %s"
    if op== "sll":
        return template % "simd_and(sisd_"+op+"i(r,sh),simd_const_"+str(fw)+"(("+str(2**fw-1)+"<<sh)&"+str(2**fw-1)+"));}\n"
    else:
        return template % "simd_and(sisd_"+op+"i(r,sh),simd_const_"+str(fw)+"("+str(2**fw-1)+">>sh));}\n"

def make_all_half_fieldwidth_versions (ops, fws):
    defn_list = ''
    for op in ops_immediateshift:
        for fw in fws_immediateshift:
            defn_list += make_immediateshift_defn(op, fw)
    for op in ops:
	if (op=="add" or op=="sub"):		
            for fw in fws[1]:
                defn_list += make_inline(op, fw, make_halfsize_defn(op,fw))
   	elif (op=="sll" or op=="srl" or op=="sra" or op=="rotl"):		
            for fw in fws[2]:
                defn_list += make_inline(op, fw, make_halfsize_defn(op,fw))
	elif (op=="pack"):		
            for fw in fws[3]:
                defn_list += make_inline(op, fw, make_halfsize_defn(op,fw))
	elif (op=="mergeh" or op=="mergel"):		
            for fw in fws[4]:
                defn_list += make_inline(op, fw, make_halfsize_defn(op,fw))	
	else: raise Exception("Bad operator %s" % op)
    return defn_list

def generate_and_write_versions(filename, ops, fws):
    defn_list = make_all_half_fieldwidth_versions (ops, fws)
    file = open(filename, 'w')
    file.write(defn_list)
    file.close()
    
if __name__ == "__main__":
    generate_and_write_versions("mmx_simd_basic.h", ops, fws)   


# text parts
processed_files = {}

# authors
for filename in ["AUTHORS", "CONTRIBUTORS"]:
    with open(filename, encoding="utf8") as f:
        text = ""
        for line in f:
            if filename == "AUTHORS":
                text += "// fast_float by " + line
            if filename == "CONTRIBUTORS":
                text += "// with contributions from " + line
        processed_files[filename] = text + "//\n//\n"

# licenses
for filename in ["LICENSE-MIT", "LICENSE-APACHE", "LICENSE-BOOST"]:
    lines = []
    with open(filename, encoding="utf8") as f:
        lines = f.readlines()

    # Retrieve subset required for inclusion in source
    if filename == "LICENSE-APACHE":
        lines = ["   Copyright 2021 The fast_float authors\n", *lines[179:-1]]

    text = ""
    for line in lines:
        line = line.strip()
        if len(line):
            line = "    " + line
        text += "//" + line + "\n"
    processed_files[filename] = text

# code
for filename in [
    "constexpr_feature_detect.h",
    "float_common.h",
    "fast_float.h",
    "ascii_number.h",
    "fast_table.h",
    "decimal_to_binary.h",
    "bigint.h",
    "digit_comparison.h",
    "parse_number.h",
]:
    with open("include/fast_float/" + filename, encoding="utf8") as f:
        text = ""
        for line in f:
            if line.startswith('#include "'):
                continue
            text += line
        processed_files[filename] = "\n" + text

# command line
import argparse

parser = argparse.ArgumentParser(description="Amalgamate fast_float.")
parser.add_argument(
    "--license",
    default="TRIPLE",
    choices=["DUAL", "TRIPLE", "MIT", "APACHE", "BOOST"],
    help="choose license",
)
parser.add_argument("--output", default="", help="output file (stdout if none")

args = parser.parse_args()


def license_content(license_arg):
    result = []
    if license_arg == "TRIPLE":
        result += [
            "// Licensed under the Apache License, Version 2.0, or the\n",
            "// MIT License or the Boost License. This file may not be copied,\n",
            "// modified, or distributed except according to those terms.\n",
            "//\n",
        ]
    if license_arg == "DUAL":
        result += [
            "// Licensed under the Apache License, Version 2.0, or the\n",
            "// MIT License at your option. This file may not be copied,\n",
            "// modified, or distributed except according to those terms.\n",
            "//\n",
        ]

    if license_arg in ("DUAL", "TRIPLE", "MIT"):
        result.append("// MIT License Notice\n//\n")
        result.append(processed_files["LICENSE-MIT"])
        result.append("//\n")
    if license_arg in ("DUAL", "TRIPLE", "APACHE"):
        result.append("// Apache License (Version 2.0) Notice\n//\n")
        result.append(processed_files["LICENSE-APACHE"])
        result.append("//\n")
    if license_arg in ("TRIPLE", "BOOST"):
        result.append("// BOOST License Notice\n//\n")
        result.append(processed_files["LICENSE-BOOST"])
        result.append("//\n")

    return result


text = "".join(
    [
        processed_files["AUTHORS"],
        processed_files["CONTRIBUTORS"],
        *license_content(args.license),
        processed_files["constexpr_feature_detect.h"],
        processed_files["float_common.h"],
        processed_files["fast_float.h"],
        processed_files["ascii_number.h"],
        processed_files["fast_table.h"],
        processed_files["decimal_to_binary.h"],
        processed_files["bigint.h"],
        processed_files["digit_comparison.h"],
        processed_files["parse_number.h"],
    ]
)

if args.output:
    with open(args.output, "wt", encoding="utf8") as f:
        f.write(text)
else:
    print(text)



import sys, os

filename = sys.argv[1]
want_size = int(sys.argv[2])

buff_size = 2**20
remains = want_size
with open(filename + '.tmp', 'wb') as fo:
    while remains > 0:
        with open(filename, 'rb') as fi:
            while remains > 0:
                chunk = fi.read(buff_size)
                if len(chunk) == 0:
                    break
                if len(chunk) > remains:
                    chunk = chunk[:remains]
                fo.write(chunk)
                remains -= len(chunk)

os.replace(filename + '.tmp', filename)


#!/usr/bin/python

# u8u16.py
#
# Python prototype implementation
# Robert D. Cameron
# revised August 5, 2009 - use generated UTF-8 definitions
#
#
#
#----------------------------------------------------------------------------
# 
# We use python's unlimited precision integers for unbounded bit streams.
# This permits simple logical operations on the entire stream.
# Assumption: bitstreams are little-endian (e.g., as on x86).
#
#----------------------------------------------------------------------------
# 

# Utility functions for demo purposes.  Slowwwww, but simple.


#
def readfile(filename):
	f = open(filename)
	contents = f.read()
	f.close()
	return contents


def count_leading_zeroes(strm):
	zeroes = 0
	while (strm & 0xFFFFFFFF) == 0: 
		zeroes += 32
		strm >>= 32
	while (strm & 1) == 0:
		zeroes += 1
		strm >>= 1
	return zeroes
	
#
def transpose_streams(s):
	
	b = []
	
	mask = 128
	index = 0
	while index < 8:
		current = 0
		cursor = 1
		for c in s:
			
			if isinstance(c,str):
				val = ord(c)
			else:
				val = c
			if  (val & mask != 0):
				current += cursor
			cursor <<= 1
			
		index+=1
		mask>>=1
		b.append(current)
	return b

def inverse_transpose(bitset, len):
	bytestream=""
	cursor = 1
	for i in range(0, len):
		byteval = 0
	       	for j in range(0,8):
       			if bitset[j] & cursor != 0:
       				byteval += 128 >> j
	       	bytestream += chr(byteval)
		cursor += cursor
	return bytestream
			
def filter_bytes(bytestream, delmask):
	newstream=""
	cursor = 1
	for c in bytestream:
		if delmask & cursor == 0:
			newstream += c
		cursor += cursor
	return newstream
			
def merge_bytes(stream1, stream2):
	s = ""
	for i in range(len(stream1)):
		s += stream1[i]
		s += stream2[i]
	return s

def bitstream2string(stream, lgth):
	str = ""
	for i in range(lgth):
		if stream & 1 == 1: str += '1'
		else: str += '_'
		stream >>= 1
	return str
#
# Advance all cursors by one position.
def Advance(stream):
	return stream + stream

def ShiftBack(stream):
	return stream >> 1


def u8_streams(u8bit):
	u8 = {}
	u8['unibyte'] = ~u8bit[0]
	u8['prefix'] = u8bit[0] & u8bit[1]
	u8['suffix'] = u8bit[0] & ~u8bit[1]
	u8prefix3or4 = u8['prefix'] & u8bit[2]
	u8['prefix2'] = u8['prefix'] & ~u8bit[2]
	u8['prefix3'] = u8prefix3or4 & ~u8bit[3]
	u8['prefix4'] = u8prefix3or4 & u8bit[3]
	u8['scope22'] = Advance(u8['prefix2'])
	u8['scope32'] = Advance(u8['prefix3'])
	u8['scope42'] = Advance(u8['prefix4'])
	u8['scope33'] = Advance(u8['scope32'])
	u8['scope43'] = Advance(u8['scope42'])
	u8['scope44'] = Advance(u8['scope43'])
	
	u8lastscope = u8['scope22'] | u8['scope33'] | u8['scope44']
	u8anyscope = u8lastscope | u8['scope32'] | u8['scope42'] | u8['scope43']
	
	# C0-C1 are illegal
	error_mask = u8['prefix2'] &~ (u8bit[3] | u8bit[4] | u8bit[5] | u8bit[6]) 
	
	prefix_E0ED = u8['prefix3'] &~ ((u8bit[6] | (u8bit[4] ^ u8bit[7])) | (u8bit[4] ^ u8bit[5]))
	E0ED_constraint = Advance(u8bit[5]) ^ u8bit[2]
	error_mask |= Advance(prefix_E0ED) &~ E0ED_constraint
	
	prefix_F5FF = u8['prefix4'] & (u8bit[4] | (u8bit[5] & (u8bit[6] | u8bit[7])))
	error_mask |= prefix_F5FF

	prefix_F0F4 = u8['prefix4'] &~ (u8bit[4] | u8bit[6] | u8bit[7])
	F0F4_constraint = Advance(u8bit[5]) ^ (u8bit[2] | u8bit[3])
	error_mask |= Advance(prefix_F0F4) &~ F0F4_constraint
	
	error_mask |= u8anyscope ^ u8['suffix']
	u8['error'] = error_mask
	return u8


#
# Generated version using chardeflist2py(DefinitionSet['UTF8'])
#
def u8_streams_generated(bit):  
  unibyte = (~bit[0]);
  prefix = (bit[0] & bit[1]);
  prefix2 = (prefix &~ bit[2]);
  temp1 = (bit[2] &~ bit[3]);
  prefix3 = (prefix & temp1);
  temp2 = (bit[2] & bit[3]);
  prefix4 = (prefix & temp2);
  suffix = (bit[0] &~ bit[1]);
  temp3 = (bit[2] | bit[3]);
  temp4 = (prefix &~ temp3);
  temp5 = (bit[4] | bit[5]);
  temp6 = (temp5 | bit[6]);
  temp7 = (temp4 &~ temp6);
  temp8 = (bit[6] | bit[7]);
  temp9 = (bit[5] & temp8);
  temp10 = (bit[4] | temp9);
  temp11 = (prefix4 & temp10);
  badprefix = (temp7 | temp11);
  temp12 = (temp5 | temp8);
  xE0 = (prefix3 &~ temp12);
  temp13 = (bit[4] & bit[5]);
  temp14 = (bit[7] &~ bit[6]);
  temp15 = (temp13 & temp14);
  xED = (prefix3 & temp15);
  xF0 = (prefix4 &~ temp12);
  temp16 = (bit[5] &~ bit[4]);
  temp17 = (temp16 &~ temp8);
  xF4 = (prefix4 & temp17);
  xA0_xBF = (suffix & bit[2]);
  x80_x9F = (suffix &~ bit[2]);
  x90_xBF = (suffix & temp3);
  x80_x8F = (suffix &~ temp3);

#
# End of generated code
  u8 = {}
  u8['unibyte'] = unibyte
  u8['prefix'] = prefix
  u8['suffix'] = suffix
  u8['prefix2'] = prefix2
  u8['prefix3'] = prefix3
  u8['prefix4'] = prefix4
  u8['scope22'] = Advance(u8['prefix2'])
  u8['scope32'] = Advance(u8['prefix3'])
  u8['scope42'] = Advance(u8['prefix4'])
  u8['scope33'] = Advance(u8['scope32'])
  u8['scope43'] = Advance(u8['scope42'])
  u8['scope44'] = Advance(u8['scope43'])
	
  u8lastscope = u8['scope22'] | u8['scope33'] | u8['scope44']
  u8anyscope = u8lastscope | u8['scope32'] | u8['scope42'] | u8['scope43']
	
  # C0-C1 and F5-FF are illegal
  error_mask = badprefix
	
  error_mask |= Advance(xE0) & x80_x9F
  error_mask |= Advance(xED) & xA0_xBF
  error_mask |= Advance(xF0) & x80_x8F
  error_mask |= Advance(xF4) & x90_xBF
	
  error_mask |= u8anyscope ^ u8['suffix']
  u8['error'] = error_mask
  return u8


#
# The following calculation of UTF-16 bit streams is consistent
# with the original u8u16, calculating streams at u8scope42 and
# u8scope44 positions.
#

def u16_streams(u8, u8bit):
	u16hi = [0,0,0,0,0,0,0,0]
	u16lo = [0,0,0,0,0,0,0,0]
	
	u8lastscope = u8['scope22'] | u8['scope33'] | u8['scope44']
	u8lastbyte = u8['unibyte'] | u8lastscope
	u16lo[2] = u8lastbyte & u8bit[2]
	u16lo[3] = u8lastbyte & u8bit[3]
	u16lo[4] = u8lastbyte & u8bit[4]
	u16lo[5] = u8lastbyte & u8bit[5]
	u16lo[6] = u8lastbyte & u8bit[6]
	u16lo[7] = u8lastbyte & u8bit[7]
	u16lo[1] = (u8['unibyte'] & u8bit[1]) | (u8lastscope & Advance(u8bit[7]))
	u16lo[0] = u8lastscope & Advance(u8bit[6])
	
	u16hi[5] = u8lastscope & Advance(u8bit[3])
	u16hi[6] = u8lastscope & Advance(u8bit[4])
	u16hi[7] = u8lastscope & Advance(u8bit[5])
	u16hi[0] = u8['scope33'] & Advance(Advance(u8bit[4]))
	u16hi[1] = u8['scope33'] & Advance(Advance(u8bit[5]))
	u16hi[2] = u8['scope33'] & Advance(Advance(u8bit[6]))
	u16hi[3] = u8['scope33'] & Advance(Advance(u8bit[7]))
	u16hi[4] = u8['scope33'] & Advance(u8bit[2])

	u8surrogate = u8['scope42'] | u8['scope44']
	u16hi[0] = u16hi[0] | u8surrogate	
	u16hi[1] = u16hi[1] | u8surrogate	
	u16hi[3] = u16hi[3] | u8surrogate	
	u16hi[4] = u16hi[4] | u8surrogate	
	u16hi[5] = u16hi[5] | u8['scope44']

	s42lo1 = ~u8bit[3] # subtract 1 
	u16lo[1] = u16lo[1] | (u8['scope42'] & s42lo1)
	s42lo0 = u8bit[2] ^ s42lo1 # borrow *
	u16lo[0] = u16lo[0] | (u8['scope42'] & s42lo0)
	borrow1 = s42lo1 & ~u8bit[2]
	s42hi7 = Advance(u8bit[7]) ^ borrow1
	u16hi[7]= u16hi[7] | (u8['scope42'] & s42hi7)
	borrow2 = borrow1 & ~Advance(u8bit[7])
	s42hi6 = Advance(u8bit[6]) ^ borrow2
	u16hi[6] = u16hi[6] | (u8['scope42'] & s42hi6)

	u16lo[2] = u16lo[2] | (u8['scope42'] & u8bit[4])
	u16lo[3] = u16lo[3] | (u8['scope42'] & u8bit[5])
	u16lo[4] = u16lo[4] | (u8['scope42'] & u8bit[6])
	u16lo[5] = u16lo[5] | (u8['scope42'] & u8bit[7])
	u16lo[6] = u16lo[6] | (u8['scope42'] & ShiftBack(u8bit[2]))
	u16lo[7] = u16lo[7] | (u8['scope42'] & ShiftBack(u8bit[3]))

	delmask = u8['prefix'] | u8['scope32'] | u8['scope43']
	return (u16hi, u16lo, delmask)




#
# The following calculation of UTF-16 bit streams uses the 
# u8scope43 position rather than the u8scope42 position for
# the bits of the first UTF-16 code unit of a surrogate pair.
# This requires more shifting than with the use of u8scope42,
# but has the advantage that all shifts are in the forward
# direction only and can hence be implemented using addition
# on little-endian architecture.
#

def u16_streams_fwdonly(u8, u8bit):
	u16hi = [0,0,0,0,0,0,0,0]
	u16lo = [0,0,0,0,0,0,0,0]
	
	u8lastscope = u8['scope22'] | u8['scope33'] | u8['scope44']
	u8lastbyte = u8['unibyte'] | u8lastscope
	u16lo[2] = u8lastbyte & u8bit[2]
	u16lo[3] = u8lastbyte & u8bit[3]
	u16lo[4] = u8lastbyte & u8bit[4]
	u16lo[5] = u8lastbyte & u8bit[5]
	u16lo[6] = u8lastbyte & u8bit[6]
	u16lo[7] = u8lastbyte & u8bit[7]
	u16lo[1] = (u8['unibyte'] & u8bit[1]) | (u8lastscope & Advance(u8bit[7]))
	u16lo[0] = u8lastscope & Advance(u8bit[6])
	
	u16hi[5] = u8lastscope & Advance(u8bit[3])
	u16hi[6] = u8lastscope & Advance(u8bit[4])
	u16hi[7] = u8lastscope & Advance(u8bit[5])
	u16hi[0] = u8['scope33'] & Advance(Advance(u8bit[4]))
	u16hi[1] = u8['scope33'] & Advance(Advance(u8bit[5]))
	u16hi[2] = u8['scope33'] & Advance(Advance(u8bit[6]))
	u16hi[3] = u8['scope33'] & Advance(Advance(u8bit[7]))
	u16hi[4] = u8['scope33'] & Advance(u8bit[2])

	u8surrogate = u8['scope43'] | u8['scope44']
	u16hi[0] = u16hi[0] | u8surrogate	
	u16hi[1] = u16hi[1] | u8surrogate	
	u16hi[3] = u16hi[3] | u8surrogate	
	u16hi[4] = u16hi[4] | u8surrogate	
	u16hi[5] = u16hi[5] | u8['scope44']


	s42lo1 = ~u8bit[3] # subtract 1 
	u16lo[1] = u16lo[1] | (u8['scope43'] & Advance(s42lo1))
	s42lo0 = u8bit[2] ^ s42lo1 # borrow *
	u16lo[0] = u16lo[0] | (u8['scope43'] & Advance(s42lo0))
	borrow1 = s42lo1 & ~u8bit[2]
	advance_bit7 = Advance(u8bit[7])
	s42hi7 = advance_bit7 ^ borrow1
	u16hi[7]= u16hi[7] | (u8['scope43'] & Advance(s42hi7))
	borrow2 = borrow1 & ~advance_bit7
	s42hi6 = Advance(u8bit[6]) ^ borrow2
	u16hi[6] = u16hi[6] | (u8['scope43'] & Advance(s42hi6))

	u16lo[2] = u16lo[2] | (u8['scope43'] & Advance(u8bit[4]))
	u16lo[3] = u16lo[3] | (u8['scope43'] & Advance(u8bit[5]))
	u16lo[4] = u16lo[4] | (u8['scope43'] & Advance(u8bit[6]))
	u16lo[5] = u16lo[5] | (u8['scope43'] & Advance(u8bit[7]))
	u16lo[6] = u16lo[6] | (u8['scope43'] & u8bit[2])
	u16lo[7] = u16lo[7] | (u8['scope43'] & u8bit[3])

	delmask = u8['prefix'] | u8['scope32'] | u8['scope42']
	return (u16hi, u16lo, delmask)



# 
# Messages to duplicate u8u16 error reporting.
#
def IllegalSequenceMessage(pos):
	return "Illegal UTF-8 sequence at position %i in source.\n" % pos

def IncompleteSequenceMessage(pos):
	return "EOF with incomplete UTF-8 sequence at position %i in source.\n" % pos


import sys
def main():

	if len(sys.argv) < 2:
		sys.stderr.write("Usage: u8u16.py u8file [u16file]\n")
		exit
	if len(sys.argv) == 3:
		outfile = open(sys.argv[2],"w")
	else: outfile = sys.stdout
	u8data = readfile(sys.argv[1])
	u8len = len(u8data)
	u8bit = transpose_streams(u8data)
#	u8 = u8_streams(u8bit)
	u8 = u8_streams_generated(u8bit)
	if u8['error'] != 0:
		err_pos = count_leading_zeroes(u8['error'])
		at_EOF = err_pos == len(u8data)
		if (err_pos >= 1) and ord(u8data[err_pos-1]) >= 0xC0: err_pos -= 1
		elif err_pos >= 2 and ord(u8data[err_pos-2]) >= 0xE0: err_pos -= 2
		elif err_pos >= 3 and ord(u8data[err_pos-3]) >= 0xF0: err_pos -= 3	
		if at_EOF:
			sys.stderr.write(IncompleteSequenceMessage(err_pos))
		else:
			sys.stderr.write(IllegalSequenceMessage(err_pos))
		u8len = err_pos
# 	Originally, we used the u16_streams version.
#	(u16hi, u16lo, delmask) = u16_streams(u8, u8bit)
	(u16hi, u16lo, delmask) = u16_streams_fwdonly(u8, u8bit)
	U16H = filter_bytes(inverse_transpose(u16hi, u8len), delmask)
	U16L = filter_bytes(inverse_transpose(u16lo, u8len), delmask)
	U16final = merge_bytes(U16H, U16L)
	outfile.write(U16final)
	outfile.close()
		
if __name__ == "__main__": main()





#!/usr/bin/env python
import sys
if(len(sys.argv)<2):
    print("please provide a file name")
    sys.exit(-1)
assert sys.version_info >= (3, 0)
filename = sys.argv[1]
counts=[0,0,0,0]#ascii, ascii+two, ascii+two_three, others
block_count=0
with open(filename, "rb") as file_content:
    array = file_content.read()
    maxv = max(array)
    minv = min(array)
    if(minv < 0): print("bug")
    if(maxv>=0b11110000):
        print("four bytes")
    elif(maxv>=0b11100000):
        print("three bytes")
    elif(maxv>=0b11000000):
        print("two bytes")
    else:
        print("ascii")
    counter = [0, 0, 0, 0]
    for x in array:
        if(x>=0b11110000):
          counter[3] += 1
        elif(x>=0b11100000):
          counter[2] += 1
        elif(x>=0b11000000):
          counter[1] += 1
        elif(x < 0b10000000):
          counter[0] += 1
        else:
          #we have a continuation byte
          pass
    print("ASCII: {}  2-Bytes: {}  3-Bytes: {} 4-Bytes: {}".format(*counter))
    l = sum(counter)
    counter = [c * 100.0 / l for c in counter]
    print("ASCII: {}%  2-Bytes: {}%  3-Bytes: {}% 4-Bytes: {}%".format(*counter))

